@@ -25,6 +25,7 @@ struct mTiming {
 	struct mTimingEvent* root;
 	struct mTimingEvent* reroot;
 
+	uint64_t globalCycles;
 	uint32_t masterCycles;
 	int32_t* relativeCycles;
 	int32_t* nextEvent;
@@ -32,12 +33,20 @@ struct mTiming {
 
 void mTimingInit(struct mTiming* timing, int32_t* relativeCycles, int32_t* nextEvent);
 void mTimingDeinit(struct mTiming* timing);
+
 void mTimingClear(struct mTiming* timing);
+void mTimingInterrupt(struct mTiming* timing);
+
 void mTimingSchedule(struct mTiming* timing, struct mTimingEvent*, int32_t when);
+void mTimingScheduleAbsolute(struct mTiming* timing, struct mTimingEvent*, int32_t when);
 void mTimingDeschedule(struct mTiming* timing, struct mTimingEvent*);
 bool mTimingIsScheduled(const struct mTiming* timing, const struct mTimingEvent*);
+
 int32_t mTimingTick(struct mTiming* timing, int32_t cycles);
+
 int32_t mTimingCurrentTime(const struct mTiming* timing);
+uint64_t mTimingGlobalTime(const struct mTiming* timing);
+
 int32_t mTimingNextEvent(struct mTiming* timing);
 int32_t mTimingUntil(const struct mTiming* timing, const struct mTimingEvent*);
 
