@@ -16,7 +16,7 @@
 
 #define MP2K_LOCK_MAX 8
 
-#ifdef _3DS
+#ifdef __3DS__
 #define blip_add_delta blip_add_delta_fast
 #endif
 
@@ -25,6 +25,7 @@ mLOG_DEFINE_CATEGORY(GBA_AUDIO, "GBA Audio", "gba.audio");
 const unsigned GBA_AUDIO_SAMPLES = 2048;
 const int GBA_AUDIO_VOLUME_MAX = 0x100;
 
+static const int SAMPLE_INTERVAL = GBA_ARM7TDMI_FREQUENCY / 0x4000;
 static const int CLOCKS_PER_FRAME = 0x800;
 
 static int _applyBias(struct GBAAudio* audio, int sample);
@@ -43,6 +44,7 @@ void GBAAudioInit(struct GBAAudio* audio, size_t samples) {
 	GBAudioInit(&audio->psg, 0, nr52, GB_AUDIO_GBA);
 	audio->psg.timing = &audio->p->timing;
 	audio->psg.clockRate = GBA_ARM7TDMI_FREQUENCY;
+	audio->psg.frameEvent.context = audio;
 	audio->samples = samples;
 	// Guess too large; we hang producing extra samples if we guess too low
 	blip_set_rates(audio->psg.left, GBA_ARM7TDMI_FREQUENCY, 96000);
@@ -57,6 +59,8 @@ void GBAAudioInit(struct GBAAudio* audio, size_t samples) {
 
 void GBAAudioReset(struct GBAAudio* audio) {
 	GBAudioReset(&audio->psg);
+	mTimingDeschedule(&audio->p->timing, &audio->psg.frameEvent);
+	mTimingSchedule(&audio->p->timing, &audio->psg.frameEvent, 0);
 	mTimingDeschedule(&audio->p->timing, &audio->sampleEvent);
 	mTimingSchedule(&audio->p->timing, &audio->sampleEvent, 0);
 	audio->chA.dmaSource = 1;
@@ -66,18 +70,22 @@ void GBAAudioReset(struct GBAAudio* audio) {
 	audio->chA.internalSample = 0;
 	audio->chA.internalRemaining = 0;
 	memset(audio->chA.fifo, 0, sizeof(audio->chA.fifo));
-	audio->chA.sample = 0;
 	audio->chB.fifoWrite = 0;
 	audio->chB.fifoRead = 0;
 	audio->chB.internalSample = 0;
 	audio->chB.internalRemaining = 0;
 	memset(audio->chB.fifo, 0, sizeof(audio->chB.fifo));
-	audio->chB.sample = 0;
-	audio->sampleRate = 0x8000;
+	int i;
+	for (i = 0; i < 8; ++i) {
+		audio->chA.samples[i] = 0;
+		audio->chB.samples[i] = 0;
+	}
 	audio->soundbias = 0x200;
 	audio->volume = 0;
 	audio->volumeChA = false;
 	audio->volumeChB = false;
+	audio->lastSample = 0;
+	audio->sampleIndex = 0;
 	audio->chARight = false;
 	audio->chALeft = false;
 	audio->chATimer = false;
@@ -85,7 +93,7 @@ void GBAAudioReset(struct GBAAudio* audio) {
 	audio->chBLeft = false;
 	audio->chBTimer = false;
 	audio->enable = false;
-	audio->sampleInterval = GBA_ARM7TDMI_FREQUENCY / audio->sampleRate;
+	audio->sampleInterval = GBA_ARM7TDMI_FREQUENCY / 0x8000;
 	audio->psg.sampleInterval = audio->sampleInterval;
 
 	blip_clear(audio->psg.left);
@@ -137,56 +145,67 @@ void GBAAudioScheduleFifoDma(struct GBAAudio* audio, int number, struct GBADMA*
 }
 
 void GBAAudioWriteSOUND1CNT_LO(struct GBAAudio* audio, uint16_t value) {
+	GBAAudioSample(audio, mTimingCurrentTime(&audio->p->timing));
 	GBAudioWriteNR10(&audio->psg, value);
 }
 
 void GBAAudioWriteSOUND1CNT_HI(struct GBAAudio* audio, uint16_t value) {
+	GBAAudioSample(audio, mTimingCurrentTime(&audio->p->timing));
 	GBAudioWriteNR11(&audio->psg, value);
 	GBAudioWriteNR12(&audio->psg, value >> 8);
 }
 
 void GBAAudioWriteSOUND1CNT_X(struct GBAAudio* audio, uint16_t value) {
+	GBAAudioSample(audio, mTimingCurrentTime(&audio->p->timing));
 	GBAudioWriteNR13(&audio->psg, value);
 	GBAudioWriteNR14(&audio->psg, value >> 8);
 }
 
 void GBAAudioWriteSOUND2CNT_LO(struct GBAAudio* audio, uint16_t value) {
+	GBAAudioSample(audio, mTimingCurrentTime(&audio->p->timing));
 	GBAudioWriteNR21(&audio->psg, value);
 	GBAudioWriteNR22(&audio->psg, value >> 8);
 }
 
 void GBAAudioWriteSOUND2CNT_HI(struct GBAAudio* audio, uint16_t value) {
+	GBAAudioSample(audio, mTimingCurrentTime(&audio->p->timing));
 	GBAudioWriteNR23(&audio->psg, value);
 	GBAudioWriteNR24(&audio->psg, value >> 8);
 }
 
 void GBAAudioWriteSOUND3CNT_LO(struct GBAAudio* audio, uint16_t value) {
+	GBAAudioSample(audio, mTimingCurrentTime(&audio->p->timing));
 	audio->psg.ch3.size = GBAudioRegisterBankGetSize(value);
 	audio->psg.ch3.bank = GBAudioRegisterBankGetBank(value);
 	GBAudioWriteNR30(&audio->psg, value);
 }
 
 void GBAAudioWriteSOUND3CNT_HI(struct GBAAudio* audio, uint16_t value) {
+	GBAAudioSample(audio, mTimingCurrentTime(&audio->p->timing));
 	GBAudioWriteNR31(&audio->psg, value);
 	audio->psg.ch3.volume = GBAudioRegisterBankVolumeGetVolumeGBA(value >> 8);
 }
 
 void GBAAudioWriteSOUND3CNT_X(struct GBAAudio* audio, uint16_t value) {
+	GBAAudioSample(audio, mTimingCurrentTime(&audio->p->timing));
 	GBAudioWriteNR33(&audio->psg, value);
 	GBAudioWriteNR34(&audio->psg, value >> 8);
 }
 
 void GBAAudioWriteSOUND4CNT_LO(struct GBAAudio* audio, uint16_t value) {
+	GBAAudioSample(audio, mTimingCurrentTime(&audio->p->timing));
 	GBAudioWriteNR41(&audio->psg, value);
 	GBAudioWriteNR42(&audio->psg, value >> 8);
 }
 
 void GBAAudioWriteSOUND4CNT_HI(struct GBAAudio* audio, uint16_t value) {
+	GBAAudioSample(audio, mTimingCurrentTime(&audio->p->timing));
 	GBAudioWriteNR43(&audio->psg, value);
 	GBAudioWriteNR44(&audio->psg, value >> 8);
 }
 
 void GBAAudioWriteSOUNDCNT_LO(struct GBAAudio* audio, uint16_t value) {
+	GBAAudioSample(audio, mTimingCurrentTime(&audio->p->timing));
 	GBAudioWriteNR50(&audio->psg, value);
 	GBAudioWriteNR51(&audio->psg, value >> 8);
 }
@@ -218,6 +237,11 @@ void GBAAudioWriteSOUNDCNT_X(struct GBAAudio* audio, uint16_t value) {
 
 void GBAAudioWriteSOUNDBIAS(struct GBAAudio* audio, uint16_t value) {
 	audio->soundbias = value;
+	int32_t oldSampleInterval = audio->sampleInterval;
+	audio->sampleInterval = 0x200 >> GBARegisterSOUNDBIASGetResolution(value);
+	if (oldSampleInterval != audio->sampleInterval && audio->p->stream && audio->p->stream->audioRateChanged) {
+		audio->p->stream->audioRateChanged(audio->p->stream, GBA_ARM7TDMI_FREQUENCY / audio->sampleInterval);
+	}
 }
 
 void GBAAudioWriteWaveRAM(struct GBAAudio* audio, int address, uint32_t value) {
@@ -229,6 +253,7 @@ void GBAAudioWriteWaveRAM(struct GBAAudio* audio, int address, uint32_t value) {
 		bank = 1;
 	}
 
+	GBAudioRun(&audio->psg, mTimingCurrentTime(audio->psg.timing), 0x4);
 	audio->psg.ch3.wavedata32[address | (bank * 4)] = value;
 }
 
@@ -241,6 +266,7 @@ uint32_t GBAAudioReadWaveRAM(struct GBAAudio* audio, int address) {
 		bank = 1;
 	}
 
+	GBAudioRun(&audio->psg, mTimingCurrentTime(audio->psg.timing), 0x4);
 	return audio->psg.ch3.wavedata32[address | (bank * 4)];
 }
 
@@ -297,7 +323,14 @@ void GBAAudioSampleFIFO(struct GBAAudio* audio, int fifoId, int32_t cycles) {
 			channel->fifoRead = 0;
 		}
 	}
-	channel->sample = channel->internalSample;
+	int32_t until = mTimingUntil(&audio->p->timing, &audio->sampleEvent) - 1;
+	int bits = 2 << GBARegisterSOUNDBIASGetResolution(audio->soundbias);
+	until += 1 << (9 - GBARegisterSOUNDBIASGetResolution(audio->soundbias));
+	until >>= 9 - GBARegisterSOUNDBIASGetResolution(audio->soundbias);
+	int i;
+	for (i = bits - until; i < bits; ++i) {
+		channel->samples[i] = channel->internalSample;
+	}
 	if (channel->internalRemaining) {
 		channel->internalSample >>= 8;
 		--channel->internalRemaining;
@@ -314,61 +347,91 @@ static int _applyBias(struct GBAAudio* audio, int sample) {
 	return ((sample - GBARegisterSOUNDBIASGetBias(audio->soundbias)) * audio->masterVolume * 3) >> 4;
 }
 
+void GBAAudioSample(struct GBAAudio* audio, int32_t timestamp) {
+	timestamp -= audio->lastSample;
+	timestamp -= audio->sampleIndex * audio->sampleInterval; // TODO: This can break if the interval changes between samples
+
+	int maxSample = 2 << GBARegisterSOUNDBIASGetResolution(audio->soundbias);
+	int sample;
+	for (sample = audio->sampleIndex; timestamp >= audio->sampleInterval && sample < maxSample; ++sample, timestamp -= audio->sampleInterval) {
+		int16_t sampleLeft = 0;
+		int16_t sampleRight = 0;
+		int psgShift = 4 - audio->volume;
+		GBAudioRun(&audio->psg, sample * audio->sampleInterval + audio->lastSample, 0xF);
+		GBAudioSamplePSG(&audio->psg, &sampleLeft, &sampleRight);
+		sampleLeft >>= psgShift;
+		sampleRight >>= psgShift;
+
+		if (audio->mixer) {
+			audio->mixer->step(audio->mixer);
+		}
+		if (!audio->externalMixing) {
+			if (!audio->forceDisableChA) {
+				if (audio->chALeft) {
+					sampleLeft += (audio->chA.samples[sample] << 2) >> !audio->volumeChA;
+				}
+
+				if (audio->chARight) {
+					sampleRight += (audio->chA.samples[sample] << 2) >> !audio->volumeChA;
+				}
+			}
+
+			if (!audio->forceDisableChB) {
+				if (audio->chBLeft) {
+					sampleLeft += (audio->chB.samples[sample] << 2) >> !audio->volumeChB;
+				}
+
+				if (audio->chBRight) {
+					sampleRight += (audio->chB.samples[sample] << 2) >> !audio->volumeChB;
+				}
+			}
+		}
+
+		sampleLeft = _applyBias(audio, sampleLeft);
+		sampleRight = _applyBias(audio, sampleRight);
+		audio->currentSamples[sample].left = sampleLeft;
+		audio->currentSamples[sample].right = sampleRight;
+	}
+
+	audio->sampleIndex = sample;
+	if (sample == maxSample) {
+		audio->lastSample += SAMPLE_INTERVAL;
+		audio->sampleIndex = 0;
+	}
+}
+
 static void _sample(struct mTiming* timing, void* user, uint32_t cyclesLate) {
 	struct GBAAudio* audio = user;
-	int16_t sampleLeft = 0;
-	int16_t sampleRight = 0;
-	int psgShift = 4 - audio->volume;
-	GBAudioSamplePSG(&audio->psg, &sampleLeft, &sampleRight);
-	sampleLeft >>= psgShift;
-	sampleRight >>= psgShift;
+	GBAAudioSample(audio, mTimingCurrentTime(&audio->p->timing) - cyclesLate);
 
-	if (audio->mixer) {
-		audio->mixer->step(audio->mixer);
-	}
-	if (!audio->externalMixing) {
-		if (!audio->forceDisableChA) {
-			if (audio->chALeft) {
-				sampleLeft += (audio->chA.sample << 2) >> !audio->volumeChA;
-			}
-
-			if (audio->chARight) {
-				sampleRight += (audio->chA.sample << 2) >> !audio->volumeChA;
-			}
-		}
-
-		if (!audio->forceDisableChB) {
-			if (audio->chBLeft) {
-				sampleLeft += (audio->chB.sample << 2) >> !audio->volumeChB;
-			}
-
-			if (audio->chBRight) {
-				sampleRight += (audio->chB.sample << 2) >> !audio->volumeChB;
-			}
-		}
-	}
-
-	sampleLeft = _applyBias(audio, sampleLeft);
-	sampleRight = _applyBias(audio, sampleRight);
+	int samples = 2 << GBARegisterSOUNDBIASGetResolution(audio->soundbias);
+	memset(audio->chA.samples, audio->chA.samples[samples - 1], sizeof(audio->chA.samples));
+	memset(audio->chB.samples, audio->chB.samples[samples - 1], sizeof(audio->chB.samples));
 
 	mCoreSyncLockAudio(audio->p->sync);
 	unsigned produced;
-	if ((size_t) blip_samples_avail(audio->psg.left) < audio->samples) {
-		blip_add_delta(audio->psg.left, audio->clock, sampleLeft - audio->lastLeft);
-		blip_add_delta(audio->psg.right, audio->clock, sampleRight - audio->lastRight);
-		audio->lastLeft = sampleLeft;
-		audio->lastRight = sampleRight;
-		audio->clock += audio->sampleInterval;
-		if (audio->clock >= CLOCKS_PER_FRAME) {
-			blip_end_frame(audio->psg.left, CLOCKS_PER_FRAME);
-			blip_end_frame(audio->psg.right, CLOCKS_PER_FRAME);
-			audio->clock -= CLOCKS_PER_FRAME;
+	int i;
+	for (i = 0; i < samples; ++i) {
+		int16_t sampleLeft = audio->currentSamples[i].left;
+		int16_t sampleRight = audio->currentSamples[i].right;
+		if ((size_t) blip_samples_avail(audio->psg.left) < audio->samples) {
+			blip_add_delta(audio->psg.left, audio->clock, sampleLeft - audio->lastLeft);
+			blip_add_delta(audio->psg.right, audio->clock, sampleRight - audio->lastRight);
+			audio->lastLeft = sampleLeft;
+			audio->lastRight = sampleRight;
+			audio->clock += audio->sampleInterval;
+			if (audio->clock >= CLOCKS_PER_FRAME) {
+				blip_end_frame(audio->psg.left, CLOCKS_PER_FRAME);
+				blip_end_frame(audio->psg.right, CLOCKS_PER_FRAME);
+				audio->clock -= CLOCKS_PER_FRAME;
+			}
+		}
+
+		if (audio->p->stream && audio->p->stream->postAudioFrame) {
+			audio->p->stream->postAudioFrame(audio->p->stream, sampleLeft, sampleRight);
 		}
 	}
 	produced = blip_samples_avail(audio->psg.left);
-	if (audio->p->stream && audio->p->stream->postAudioFrame) {
-		audio->p->stream->postAudioFrame(audio->p->stream, sampleLeft, sampleRight);
-	}
 	bool wait = produced >= audio->samples;
 	if (!mCoreSyncProduceAudio(audio->p->sync, audio->psg.left, audio->samples)) {
 		// Interrupted
@@ -379,7 +442,7 @@ static void _sample(struct mTiming* timing, void* user, uint32_t cyclesLate) {
 		audio->p->stream->postAudioBuffer(audio->p->stream, audio->psg.left, audio->psg.right);
 	}
 
-	mTimingSchedule(timing, &audio->sampleEvent, audio->sampleInterval - cyclesLate);
+	mTimingSchedule(timing, &audio->sampleEvent, SAMPLE_INTERVAL - cyclesLate);
 }
 
 void GBAAudioSerialize(const struct GBAAudio* audio, struct GBASerializedState* state) {
@@ -387,12 +450,18 @@ void GBAAudioSerialize(const struct GBAAudio* audio, struct GBASerializedState*
 
 	STORE_32(audio->chA.internalSample, 0, &state->audio.internalA);
 	STORE_32(audio->chB.internalSample, 0, &state->audio.internalB);
-	state->audio.sampleA = audio->chA.sample;
-	state->audio.sampleB = audio->chB.sample;
+	memcpy(state->samples.chA, audio->chA.samples, sizeof(audio->chA.samples));
+	memcpy(state->samples.chB, audio->chB.samples, sizeof(audio->chB.samples));
+
+	size_t i;
+	for (i = 0; i < GBA_MAX_SAMPLES; ++i) {
+		STORE_16(audio->currentSamples[i].left, 0, &state->currentSamples[i].left);
+		STORE_16(audio->currentSamples[i].right, 0, &state->currentSamples[i].right);
+	}
+	STORE_32(audio->lastSample, 0, &state->audio.lastSample);
 
 	int readA = audio->chA.fifoRead;
 	int readB = audio->chB.fifoRead;
-	size_t i;
 	for (i = 0; i < GBA_AUDIO_FIFO_SIZE; ++i) {
 		STORE_32(audio->chA.fifo[readA], i << 2, state->audio.fifoA);
 		STORE_32(audio->chB.fifo[readB], i << 2, state->audio.fifoB);
@@ -426,6 +495,11 @@ void GBAAudioSerialize(const struct GBAAudio* audio, struct GBASerializedState*
 	flags = GBASerializedAudioFlagsSetFIFOInternalSamplesA(flags, audio->chA.internalRemaining);
 	flags = GBASerializedAudioFlagsSetFIFOInternalSamplesB(flags, audio->chB.internalRemaining);
 	STORE_16(flags, 0, &state->audio.gbaFlags);
+
+	GBASerializedAudioFlags2 flags2 = 0;
+	flags2 = GBASerializedAudioFlags2SetSampleIndex(flags2, audio->sampleIndex);
+	STORE_32(flags2, 0, &state->audio.gbaFlags2);
+
 	STORE_32(audio->sampleEvent.when - mTimingCurrentTime(&audio->p->timing), 0, &state->audio.nextSample);
 }
 
@@ -434,12 +508,18 @@ void GBAAudioDeserialize(struct GBAAudio* audio, const struct GBASerializedState
 
 	LOAD_32(audio->chA.internalSample, 0, &state->audio.internalA);
 	LOAD_32(audio->chB.internalSample, 0, &state->audio.internalB);
-	audio->chA.sample = state->audio.sampleA;
-	audio->chB.sample = state->audio.sampleB;
+	memcpy(audio->chA.samples, state->samples.chA, sizeof(audio->chA.samples));
+	memcpy(audio->chB.samples, state->samples.chB, sizeof(audio->chB.samples));
+
+	size_t i;
+	for (i = 0; i < GBA_MAX_SAMPLES; ++i) {
+		LOAD_16(audio->currentSamples[i].left, 0, &state->currentSamples[i].left);
+		LOAD_16(audio->currentSamples[i].right, 0, &state->currentSamples[i].right);
+	}
+	LOAD_32(audio->lastSample, 0, &state->audio.lastSample);
 
 	int readA = 0;
 	int readB = 0;
-	size_t i;
 	for (i = 0; i < GBA_AUDIO_FIFO_SIZE; ++i) {
 		LOAD_32(audio->chA.fifo[readA], i << 2, state->audio.fifoA);
 		LOAD_32(audio->chB.fifo[readB], i << 2, state->audio.fifoB);
@@ -456,8 +536,15 @@ void GBAAudioDeserialize(struct GBAAudio* audio, const struct GBASerializedState
 	audio->chA.internalRemaining = GBASerializedAudioFlagsGetFIFOInternalSamplesA(flags);
 	audio->chB.internalRemaining = GBASerializedAudioFlagsGetFIFOInternalSamplesB(flags);
 
+	GBASerializedAudioFlags2 flags2;
+	LOAD_32(flags2, 0, &state->audio.gbaFlags2);
+	audio->sampleIndex = GBASerializedAudioFlags2GetSampleIndex(flags2);
+
 	uint32_t when;
 	LOAD_32(when, 0, &state->audio.nextSample);
+	if (state->versionMagic < 0x01000007) {
+		audio->lastSample = when - SAMPLE_INTERVAL;
+	}
 	mTimingSchedule(&audio->p->timing, &audio->sampleEvent, when);
 }
 
