@@ -5,6 +5,7 @@
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
 #include "GDBWindow.h"
 
+#include <QButtonGroup>
 #include <QGridLayout>
 #include <QGroupBox>
 #include <QLabel>
@@ -12,6 +13,7 @@
 #include <QMessageBox>
 #include <QPushButton>
 #include <QHBoxLayout>
+#include <QRadioButton>
 #include <QVBoxLayout>
 
 #include "GDBController.h"
@@ -47,6 +49,30 @@ GDBWindow::GDBWindow(GDBController* controller, QWidget* parent)
 	connect(m_bindAddressEdit, &QLineEdit::textChanged, this, &GDBWindow::bindAddressChanged);
 	settingsGrid->addWidget(m_bindAddressEdit, 1, 1, Qt::AlignLeft);
 
+	QGroupBox* watchpointsSettings = new QGroupBox(tr("Write watchpoints behavior"));
+	mainSegment->addWidget(watchpointsSettings);
+
+	QVBoxLayout* watchpointsSettingsLayout = new QVBoxLayout;
+	watchpointsSettings->setLayout(watchpointsSettingsLayout);
+
+	QButtonGroup* watchpointsButtonGroup = new QButtonGroup(watchpointsSettings);
+	connect(watchpointsButtonGroup, qOverload<QAbstractButton*>(&QButtonGroup::buttonClicked), controller, [watchpointsButtonGroup, controller](QAbstractButton* button) {
+		controller->setWatchpointsBehavior(watchpointsButtonGroup->id(button));
+	});
+
+	m_watchpointsStandardRadio = new QRadioButton(tr("Standard GDB"), watchpointsSettings);
+	m_watchpointsStandardRadio->setChecked(true);
+	watchpointsButtonGroup->addButton(m_watchpointsStandardRadio, GDB_WATCHPOINT_STANDARD_LOGIC);
+	watchpointsSettingsLayout->addWidget(m_watchpointsStandardRadio);
+
+	m_watchpointsOverrideLogicRadio = new QRadioButton(tr("Internal change detection"), watchpointsSettings);
+	watchpointsButtonGroup->addButton(m_watchpointsOverrideLogicRadio, GDB_WATCHPOINT_OVERRIDE_LOGIC);
+	watchpointsSettingsLayout->addWidget(m_watchpointsOverrideLogicRadio);
+
+	m_watchpointsOverrideLogicAnyWriteRadio = new QRadioButton(tr("Break on all writes"), watchpointsSettings);
+	watchpointsButtonGroup->addButton(m_watchpointsOverrideLogicAnyWriteRadio, GDB_WATCHPOINT_OVERRIDE_LOGIC_ANY_WRITE);
+	watchpointsSettingsLayout->addWidget(m_watchpointsOverrideLogicAnyWriteRadio);
+
 	QHBoxLayout* buttons = new QHBoxLayout;
 
 	m_startStopButton = new QPushButton;
@@ -92,6 +118,9 @@ void GDBWindow::bindAddressChanged(const QString& bindAddressString) {
 void GDBWindow::started() {
 	m_portEdit->setEnabled(false);
 	m_bindAddressEdit->setEnabled(false);
+	m_watchpointsStandardRadio->setEnabled(false);
+	m_watchpointsOverrideLogicRadio->setEnabled(false);
+	m_watchpointsOverrideLogicAnyWriteRadio->setEnabled(false);
 	m_startStopButton->setText(tr("Stop"));
 	m_breakButton->setEnabled(true);
 	disconnect(m_startStopButton, &QAbstractButton::clicked, m_gdbController, &GDBController::listen);
@@ -102,6 +131,9 @@ void GDBWindow::started() {
 void GDBWindow::stopped() {
 	m_portEdit->setEnabled(true);
 	m_bindAddressEdit->setEnabled(true);
+	m_watchpointsStandardRadio->setEnabled(true);
+	m_watchpointsOverrideLogicRadio->setEnabled(true);
+	m_watchpointsOverrideLogicAnyWriteRadio->setEnabled(true);
 	m_startStopButton->setText(tr("Start"));
 	m_breakButton->setEnabled(false);
 	disconnect(m_startStopButton, &QAbstractButton::clicked, m_gdbController, &DebuggerController::detach);
