@@ -32,6 +32,10 @@ public:
 
 	void setInputController(InputController* controller);
 	void setMode(LoadSave mode);
+	void setBackground(const QPixmap& pixmap) { m_background = pixmap; }
+	void setDimensions(const QSize& dims) { m_dims = dims; }
+	void setLockIntegerScaling(bool lockIntegerScaling) { m_lockIntegerScaling = lockIntegerScaling; }
+	void setLockAspectRatio(bool lockApsectRatio) { m_lockAspectRatio = lockApsectRatio; }
 
 signals:
 	void closed();
@@ -54,6 +58,11 @@ private:
 
 	int m_currentFocus;
 	QPixmap m_currentImage;
+	QPixmap m_background;
+
+	QSize m_dims;
+	bool m_lockAspectRatio;
+	bool m_lockIntegerScaling;
 };
 
 }
